#!/usr/bin/env python

"""Utility module with functions to provide paths to application files.

.. module:: paths
   :platform: Windows
   :synopsis: Provides application paths

.. moduleauthor:: Mathew Topper <mathew.topper@dataonlygreater.com>

"""

import inspect
import os
import shutil
import sys
from pathlib import Path
from typing import Optional

from .appdirs import site_data_dir, system, user_data_dir


class ModPath(Path):
    """Path relative to a given module name's parent directory"""

    def __new__(cls, module_name, *paths):
        mod_handle = sys.modules[module_name]
        dir_path = module_dir(mod_handle)
        return Path(dir_path, *paths)


class EtcPath(Path):
    """Path relative to the Python distribution's etc directory"""

    def __new__(cls, *paths):
        def get_dir(*paths):  # pylint: disable=missing-docstring
            return os.path.abspath(os.path.join(*paths))

        exe_folder = os.path.dirname(sys.executable)

        if system == "win32":
            dir_path = get_dir(exe_folder, "etc")
        else:
            dir_path = get_dir(exe_folder, "..", "etc")

        return Path(dir_path, *paths)


class UserDataPath(Path):
    """Path relative to the user data directory for the given package."""

    def __new__(cls, package, company, *paths):
        dir_path = user_data_dir(package, company, roaming=True)
        dir_path = os.path.join(dir_path, *paths)
        return Path(dir_path)


class SiteDataPath(Path):
    """Path relative to the site data directory for the given package."""

    def __new__(cls, package, company, *paths):
        dir_path = site_data_dir(package, company)
        return Path(dir_path, *paths)


class DirectoryMap:
    """Class for moving files from one directory to another.

    Attributes:
        target_dir (pathlib.Path)
        source_dir (pathlib.Path)
        last_copy_path (str): Location of last copied configuration file.
          Defaults to None.

    """

    def __init__(self, target_dir, source_dir):
        self.target_dir: Path = target_dir
        self.source_dir: Path = source_dir
        self.last_copy_path: Optional[str] = None

        return

    def target_exists(self, file_name=None):
        """Test to see if a given target file or directory exists

        Args:
            file_name (str, optional): The name of a target file to
              check.

        Returns:
            bool: True if path exists.

        """

        if file_name is not None:
            result = (self.target_dir / file_name).is_file()

        else:
            result = self.target_dir.is_dir()

        return result

    def copy_file(
        self,
        src_name,
        dst_name=None,
        overwrite=False,
        new_ext=".new",
    ):
        """Copy a file from the source to target directory.

        Args:
            src_name (str): File name of the file to copy. It is
              assumed that this exists within the source directory.
            dst_name (str, optional): Destination file name. Defaults to the
              same as src_name.
            overwrite (bool, optional): Copy the files to target
              directory even if it already exists. Default to False.
            new_ext (str, optional): If target file exists and overwrite
              is False copy the file appending this extension. Defaults to
              ".new".

        """

        if self.source_dir is None:
            error_str = "No source directory is given."
            raise ValueError(error_str)

        if dst_name is None:
            file_name = src_name
        else:
            file_name = dst_name

        # Build the destination path
        dst_file_path = self.target_dir / file_name

        if self.target_exists(file_name) and not overwrite:
            dst_file_path = dst_file_path.with_suffix(
                dst_file_path.suffix + new_ext
            )

        # Get the source path
        src_file_path = self.source_dir / src_name

        # Prepare the destination directory
        self.target_dir.mkdir(exist_ok=True)

        # Copy the file
        shutil.copy(src_file_path, dst_file_path)

        # Punch out the dst_file_path to logging
        #        sys.stdout.write(dst_file_path)

        # Store the location of the copy directory
        self.last_copy_path = dst_file_path

        return

    def safe_copy_file(
        self,
        src_name,
        dst_name=None,
        **kwargs,
    ):
        """Copy the file to the target directory if it does not
        exist already.

        Args:
            src_name (str): File name of the file to copy. It is
              assumed that this exists within the source directory.
            dst_name (str, optional): Destination file name. Defaults to the
              same as src_name.
        """

        if dst_name is None:
            file_name = src_name
        else:
            file_name = dst_name

        # Test for existance otherwise copy the file to the user data
        # directory
        if not self.target_exists(file_name):
            self.copy_file(src_name, dst_name)

        return

    def copy_all(
        self,
        overwrite=False,
        new_ext=".new",
    ):
        """Copy all files from the source to target directory.

        Args:
            overwrite (bool, optional): Copy the files to target
              directory even if it already exists. Default to False.
            new_ext (str, optional): If target file exists and overwrite
              is False copy the file appending this extension. Defaults to
              ".new".

        """

        self._copy_all(self.copy_file, overwrite=overwrite, new_ext=new_ext)

        return

    def safe_copy_all(self):
        """Copy all files from the source to target directory if they do not
        exist.
        """

        self._copy_all(self.safe_copy_file)

        return

    def _copy_all(
        self,
        copy_method,
        overwrite=False,
        new_ext=".new",
    ):
        """Copy all files from the source to target directory.

        Args:
            copy_method: File copying method to call
            overwrite (bool, optional): Copy the files to target
              directory even if it already exists. Default to False.
            new_ext (str, optional): If target file exists and overwrite
              is False copy the file appending this extension. Defaults to
              ".new".

        """

        for src_path in self.source_dir.iterdir():
            kwargs = {"overwrite": overwrite, "new_ext": new_ext}
            copy_method(src_path.name, **kwargs)

        return


def object_path(obj):
    """Return the path of the class object.

    Args:
      object (obj): Class object.

    Returns:
        pathlib.Path: Absolute path to class source file.
    """

    return Path(inspect.getabsfile(obj.__class__))


def object_dir(obj):
    """Return the directory path containing the class object.

    Args:
      object (obj): Class object.

    Returns:
        pathlib.Path:: Absolute path to directory containing class source file.
    """

    obj_path = object_path(obj)
    return obj_path.parent


def class_path(cls):
    """Return the path of the class attribute.

    Args:
      class (cls): Class.

    Returns:
        pathlib.Path:: Absolute path to class source file.
    """

    return Path(inspect.getabsfile(cls))


def class_dir(cls):
    """Return the directory path containing the class attribute.

    Args:
      class (cls): Class.

    Returns:
        pathlib.Path:: Absolute path to directory containing class source file.
    """

    cls_path = class_path(cls)
    return cls_path.parent


def module_path(module):
    """Return the path of the modules "__file__" attribute.

    Args:
      module (module): Module object.

    Returns:
        pathlib.Path:: Absolute path to "module.__file__".

    """

    return Path(module.__file__).resolve()


def module_dir(module):
    """Returns path to the directory containing the given module.

    Args:
      module (module): Module object.

    Returns:
        pathlib.Path:: Absolute path to directory containing module.

    """

    mod_path = module_path(module)
    return mod_path.parent



# -*- coding: utf-8 -*-

"""This module provides function for configuring the application. This
includes settings for logging and helpers for user modified configuration
files.

.. module:: configuration
   :platform: Windows
   :synopsis: Provides configuration helpers

.. moduleauthor:: Mathew Topper <mathew.topper@dataonlygreater.com>

"""

# # Set up logging
# import logging

# module_logger = logging.getLogger(__name__)

# Built in modules
import logging
from logging.config import dictConfig
from pathlib import Path
from typing import Union

# External modules
import yaml
from configobj import ConfigObj
from validate import ValidateError, Validator

try:
    from yaml import CDumper as Dumper
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Dumper, Loader

# Local modules
from .paths import DirectoryMap


class Config():
    """Base class for handling configuration files.

    Attributes:
        directory (polite.paths.Directory, polite.paths.DirectoryMap)
        file_name (str): File name of the config file.

    """

    def __init__(self, directory: Union[Path, DirectoryMap], file_name):
        if isinstance(directory, DirectoryMap):
            self.target_dir = directory.target_dir
            self.directory_map = directory

        else:
            self.target_dir = directory
            self.directory_map = None

        self.config_file_name = file_name

    def get_config_path(self):
        return self.target_dir / self.config_file_name

    def copy_config(self, overwrite=False, new_ext=".new"):
        """Copy a user editable config file to the target directory.

        Args:
            overwrite (bool, optional): Copy the config files to
              target directory even if it already exists. Default to False
            new_dir (str, optional): If user_config_path exists and overwrite
              is False copy the configuration with this extension. Defaults to
              ".new".

        """

        if self.directory_map is None:
            error_str = "No source directory available."
            raise ValueError(error_str)

        self.directory_map.copy_file(
            self.config_file_name, overwrite=overwrite, new_ext=new_ext
        )

    def config_exists(self):
        result = (self.target_dir / self.config_file_name).is_file()
        return result

    @classmethod
    def make_head_foot_bar(cls, header_title, bar_width, bar_char="*"):
        """Make header and footer strings consisting of a bar of characters of
        fixed width, with a title embeded in the header bar.

        Args:
            header_title (str): The title to be placed in the header bar.
            bar_width (int): The number of characters in the header and
              footer.
            bar_char (str, optional): Character to use for the bar. Default is
              "*".

        Returns:
            tuple: Tuple containing the strings (header, footer).
        """

        # Build the header and footer
        title_space = " {} ".format(header_title)
        header = "{0:{1}^{2}}".format(title_space, bar_char[0], bar_width)
        footer = bar_char * bar_width

        return (header, footer)


class ReadINI(Config):
    """Class to prepare and read ini configuration files. This class uses
    the configobj package.

    Args:
        validation_file_name (str, optional): Name of the file to be used to
          validate the inputs to the configuration file. Defaults to
          "validation.ini".

    Attributes:
        validation_file_name (str): File name of the configobj validation file.

    """

    def __init__(
        self,
        directory,
        config_file_name="configuration.ini",
        validation_file_name=None,
    ):
        super(ReadINI, self).__init__(directory, config_file_name)
        self.validation_file_name = validation_file_name

    def get_validation_path(self):
        if self.validation_file_name is None:
            return self.target_dir
        return self.target_dir / self.validation_file_name

    def copy_config(self, overwrite=False, new_ext=".new"):
        """Copy a user editable config file to the target directory.

        Args:
            overwrite (bool, optional): Copy the config files to
              target directory even if it already exists. Default to False
            new_dir (str, optional): If user_config_path exists and overwrite
              is False copy the configuration with this extension. Defaults to
              ".new".

        """

        super(ReadINI, self).copy_config(overwrite, new_ext)

        if self.validation_file_name is None:
            return

        assert self.directory_map is not None
        self.directory_map.copy_file(
            self.validation_file_name, overwrite=overwrite, new_ext=new_ext
        )

    def config_exists(self):
        result = super(ReadINI, self).config_exists()

        if self.validation_file_name is not None:
            result = (
                result
                and (self.target_dir / self.validation_file_name).is_file()
            )

        return result

    def get_config(self, configspec_path=None):
        """Load the INI configuration file to a ConfigObj class,
        without validation, but allowing a configspec to be given.
        """

        # Get the path to the configspec file
        ini_config_path = self.get_config_path()

        # Test for existance of the file
        if not self.config_exists():
            error_str = "Expected file not found at path: {}".format(
                ini_config_path
            )
            raise IOError(error_str)

        if configspec_path is None:
            configspec = None
        else:
            configspec = str(configspec_path)

        config = ConfigObj(
            str(ini_config_path),
            configspec=configspec,
        )

        return config

    def get_valid_config(self, error_title="Config File Errors", bar_width=42):
        """Check whether the configuration file is valid, using the
        validation file specified in self.validation_file_name.

        This file is anticipated to be located with the "config" directory
        of the source code and the file name is appended to this directory
        path.

        Args:
            error_title (str, optional): The title to be placed in the header
              of the validation error messages. Default "Config File Errors".
            bar_width (int, optional): The number of characters in the header
              and footer. Default is 42.
        """

        # Get the path to the configspec file
        ini_valid_path = self.get_validation_path()
        config = self.get_config(ini_valid_path)

        validator = Validator()
        result = config.validate(validator, preserve_errors=True)

        if result != True:
            # Build the header and footer
            header, footer = self.make_head_foot_bar(error_title, bar_width)

            # Iterate through the failures in the config file
            log_msgs = ["", header]
            log_msgs.extend(self._type_fails(result))
            log_msgs.append(footer)

            log_msg = "\n".join(log_msgs)

            # module_logger.debug(log_msg)

            error_str = (
                'Configuration file "{}" failed ' "validation:\n{}"
            ).format(self.config_file_name, log_msg)
            raise ValidateError(error_str)

        return config

    def _type_fails(self, results):
        """Create strings with the specific validation errors.

        Args:
            results: The results of a ConfigObj.validate call.

        Results:
            list: A list of strings containing the validation errors.

        """

        log_lines = []

        # Iterate through the failures in the config file
        for key, value in results.items():
            if issubclass(type(value), ValidateError):
                log_str = (' - Key "{}" failed with error:\n' "   {}").format(
                    key, value
                )
                log_lines.append(log_str)

            elif value is False:
                log_str = ' - Key "{}" must be set.'.format(key)
                log_lines.append(log_str)

            elif isinstance(value, dict):
                add_logs = self._type_fails(value)
                log_lines.extend(add_logs)

        return log_lines


class ReadYAML(Config):
    """Class to read and write YAML configuration files. This class uses
    the pyyaml package.

    """

    def __init__(self, directory, yaml_file_name):
        super(ReadYAML, self).__init__(directory, yaml_file_name)

        return

    def read(self):
        """Load the YAML configuration file."""

        # Get the file path
        yaml_config_path = self.get_config_path()

        with open(yaml_config_path, "r") as conf:
            config_dict = yaml.load(conf, Loader=Loader)

        return config_dict

    def write(self, obj_to_serialise, default_flow_style=False):
        """Write the YAML configuration file."""

        # Write the file
        yaml_config_path = self.get_config_path()

        # Ensure target directory exists
        self.target_dir.mkdir(exist_ok=True)

        with open(yaml_config_path, "w") as yaml_file:
            yaml.dump(
                obj_to_serialise,
                yaml_file,
                default_flow_style=default_flow_style,
                Dumper=Dumper,
            )


class Logger(ReadYAML):
    """Class to configure and control the python logging system.

    Args:
        config_file_name (str, optional): Name of the logging config file.
          Defaults to "logging.yaml".

    Attributes:
        logging_config_file_name (str): File name of the logging config file.

    """

    def __init__(self, directory, config_file_name="logging.yaml"):
        super(Logger, self).__init__(directory, config_file_name)

    @classmethod
    def configure_logger(cls, log_config_dict):
        """Load the logging configuration file."""

        # Configure the logger
        dictConfig(log_config_dict)

    @classmethod
    def add_named_logger(cls, log_name, log_level=None, info_message=None):
        """Start a named logger.

        Args:
            log_name (str): Name of the logger.
            log_level (str, optional): Valid logging level.
            info_message (str, optional): Message to broadcast on the logger
              info channel.
        """

        named_logger = logging.getLogger(log_name)

        if log_level is not None:
            named_logger.setLevel(log_level)

        if info_message is not None:
            named_logger.info(info_message)

        return named_logger

    def __call__(self, package, level=None, info_message=None):
        # Bring up the logger
        if self.directory_map is not None:
            self.copy_config()
        log_config_dict = self.read()
        self.configure_logger(log_config_dict)
        self.add_named_logger(package, level, info_message)
